# -*- coding: utf-8 -*-
import sys

from tfutility.core.base import Command
from tfutility.core.tffile import TfFile
from tfutility.core.tfpaths import TFPaths


class ForcedRemoteSourceHandler(TFPaths, Command):
    name = "forcedremotesource"
    help = "Check if a RemoteSource was set"

    def add_arguments(self, parser):
        parser = super().add_arguments(parser)
        parser.add_argument(
            "-s", "--silent", action="store_true", help="Prevent any log output"
        )
        parser.add_argument(
            "--allow-failure",
            action="store_true",
            help="If this flag was set, the module logs only all occurences but does not exit with code 1",
        )
        return parser

    def new_decorator(self, options, block):
        file_path = block.get_tf_file().path
        if not block.id.startswith("module"):
            self.get_logger().error(
                "The decorator @forcedremotesource can only applied to modules"
            )
            sys.exit(1)

        block_content = block.content

        if not block_content.get("version"):
            self._error = True
            self.get_logger().error(
                f"Module Block had no Version Defined in {file_path}:{block.start}"
            )

        if block_content.get("source")[0] == ".":
            self._error = True
            self.get_logger().error(
                f"Module Block has no Remote Source in {file_path}:{block.start}"
            )

    def handle(self, options):
        self._error = False
        if options.silent:
            self.get_logger().setLevel(1000)

        tf_files = self.get_file_list(options.paths)

        for file in tf_files:
            file = TfFile(file)

            for block in file.get_blocks_with_decorator(self.get_name()):
                self.new_decorator(self, block)

        if not options.allow_failure and self._error:
            sys.exit(1)



# -*- coding: utf-8 -*-
import hcl2

from tfutility.core.tffile import OpendTfFile, TfBlock, TfFile


def test_filemetadata_extract_blocks_simple(mocker):
    mocker.patch("tfutility.core.tffile.TfFile.read_tf")
    fm = TfFile("")

    fm._extract_blocks(
        {
            "terraform": [
                {
                    "required_providers": [
                        {"aws": "fooo", "__start_line__": 5, "__end_line__": 5}
                    ],
                    "__start_line__": 1,
                    "__end_line__": 5,
                }
            ]
        },
        "",
    )
    targets = ["terraform.required_providers"]
    assert all([block.id in targets for block in fm._blocks])


def test_filemetadata_extract_blocks_more(mocker):
    mocker.patch("tfutility.core.tffile.TfFile.read_tf")
    fm = TfFile("")

    with open("./tests/testcode/main.tf", "r") as fobj:
        data = hcl2.load(fobj, with_meta=True)

    fm._extract_blocks(data, "")
    targets = [
        "terraform.required_providers",
        "provider.aws",
        "resource.aws_s3_bucket.example",
        "resource.aws_s3_bucket_object.object2",
        "module.testmodule_localsource.source",
        "module.testmodule_remotesource.source",
        "moved",
    ]
    print(fm._blocks)
    assert all([block.id in targets for block in fm._blocks])


def test_find_decorators_nothing_found(mocker):
    mocker.patch(
        "tfutility.core.tffile.TfFile.read_tf",
        return_value=OpendTfFile(None, ["a", "b", "c", "d"], None),
    )
    tff = TfFile("", autoparse=False)
    bw = TfBlock(tff, "foo", {"__start_line__": 5, "__end_line__": 5})
    assert bw._find_decorators() == []


def test_find_decorators_found(mocker):
    mocker.patch(
        "tfutility.core.tffile.TfFile.read_tf",
        return_value=OpendTfFile(
            None,
            ['# @bar(bar="test")', "a", '# @fooo(bar="test", party="hard")', "c", "d"],
            None,
        ),
    )
    tff = TfFile("", autoparse=False)
    nbw = TfBlock(tff, "foo", {"__start_line__": 4, "__end_line__": 5})
    x = nbw._find_decorators()
    assert x[0].get_name() == "fooo"
    assert len(x) == 1
    assert x[0].parameter("bar") == "test"
    assert x[0].parameter("party") == "hard"



# -*- coding: utf-8 -*-
import argparse


class AbstractCommand:
    """
    An abstract class definition for custom CLI Commands
    """

    def __init__(self, parent, parser: argparse.ArgumentParser):
        self._cmd = parent
        self.parser = parser

    def _init(self):
        """
        Intialize an subparser for this command
        normaly it has the name of this command
        """
        raise NotImplementedError("Please Implement an _init method")

    def get_logger(self):
        """

        :return: The logger from the main class

        """
        return self._cmd.get_logger()

    def add_arguments(self, parser: argparse.ArgumentParser):
        raise NotImplementedError("Please implement an add_arguments method")

    def get_name(self):
        if self.__class__.name is None:
            raise NotImplementedError("Please give your Command a name arg ")
        return self.name

    def get_help(self):
        if self.__class__.help is None:
            raise NotImplementedError("Please give your Command a help arg")
        return self.help

    def itsme(self, options: argparse.Namespace) -> bool:
        if options.command == self.get_name():
            return True
        return False

    def handle(self, option):
        raise NotImplementedError("Please implement this method in {self.__class__}")



# -*- coding: utf-8 -*-
import argparse
import pathlib

from tfutility.core.abstract import AbstractCommand
from tfutility.core.exp import PathIsNotValid


class TFPaths(AbstractCommand):
    """Add an CLI argument to the command to Found .tf files"""

    SEARCH_GLOB = "**/*.tf"

    def add_arguments(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        """Add argument that consumes multiple filesystem paths.

        :param parser: A ArgumentParser object. if it is None the default one (for the whole application) will be used.
        :type parser: argparse.ArgumentParser
        :return: Tthe argument parser enriched with arguments
        :rtype: argparse.ArgumentParser
        """
        parser = super(TFPaths, self).add_arguments(parser)
        parser.add_argument(
            "paths",
            nargs="+",
            type=pathlib.Path,
            help="Path to one or more TF-Files. Its also Possible to set a folder. It searches only *.tf files",
        )
        return parser

    def _get_files_from_path(self, path: pathlib.Path):
        """Resolves a given Path to a list of all matched files in it

        :param path: Path Object of target folder or file on the filesystem
        :type path: pathlib.Path
        :raises PathIsNotValid: when the Path does not exist
        :return: list
        :rtype: list[pathlib.Path]
        """
        if not path.exists():
            raise PathIsNotValid("Path does not exist")
        if path.is_dir():
            return list(path.glob(self.__class__.SEARCH_GLOB))
        return [path]

    def get_file_list(self, paths: pathlib.Path) -> list[pathlib.Path]:
        """Resolve all given cli nargs. If a given path has no file matching it logs a warning

        :param paths: list[pathlib.Path]
        :type paths: pathlib.Path
        :return: List of all found Files
        :rtype: list[pathlib.Path]
        """
        file_list = []
        for p in paths:
            file_list += self._get_files_from_path(p)

        if len(file_list) <= 0:
            self.get_logger().warning(
                "The Search-Methoud found Zero files tha tcan be used by tfutility "
            )
        return file_list



# -*- coding: utf-8 -*-
import argparse
import logging
from collections import namedtuple

from tfutility.controllers.blockdate import ImportDateHandler, MovedDateHandler
from tfutility.controllers.forcedremotesource import ForcedRemoteSourceHandler
from tfutility.controllers.sourceswap import SourceSwapHandler

HandlerInfo = namedtuple("HandlerInfo", ("handler_cls", "handler_obj", "init_parser"))

log = logging.getLogger("tfutility")


class TfUtility:
    """
    TfUtility Initial class. This class initializes all CLI arguments
    """

    # List of all Available CLI Handlers
    handlers = [
        ForcedRemoteSourceHandler,
        ImportDateHandler,
        MovedDateHandler,
        SourceSwapHandler,
    ]

    def __init__(self):
        self.commands = {}

    def _init_handlers(self, parser: argparse.ArgumentParser):
        """
        initialize all cli argparse commands

        :type parser: argparse.ArgumentParser
        :param parser: the main argparser
        """
        subparser = parser.add_subparsers(dest="command")

        for handler_cls in self.__class__.handlers:
            handler = handler_cls(self, parser)
            cmd_parser = handler._init(parser, subparser)
            handler.add_arguments(cmd_parser)
            self.commands[handler.get_name()] = HandlerInfo(
                handler_cls, handler, cmd_parser
            )

    def get_logger(self) -> logging.Logger:
        """
        get logger instance

        :return: the logger instance
        :rtype: logging.Logger
        """
        return log

    def add_arguments(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        """
        Add global arguments to the parser

        :param parser: the main argparser
        :type parser: argparse.ArgumentParser
        :return: the main argparser with attached global arguments
        """
        parser.add_argument("--verbose", "-v", action="count", default=1)
        return parser

    def _handle(self, options: argparse.Namespace):
        """
        handle the cli arguments

        :param options: the parsed cli arguments
        :type options: argparse.Namespace
        """

        # set default loglevel to Info
        options.verbose = 40 - (10 * options.verbose) if options.verbose > 0 else 0
        logging.basicConfig(
            level=options.verbose,
            format="%(asctime)s %(levelname)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

        # find command which is used by user
        for name, handler in self.commands.items():
            if handler.handler_obj.itsme(options):
                handler.handler_obj.handle(options)

    def _init_parser(self) -> argparse.ArgumentParser:
        """
        Create the main Parser with all subparsers arguments
        """
        parser = argparse.ArgumentParser()
        parser = self.add_arguments(parser)
        self._init_handlers(parser)
        return parser

    def do(self):
        """
        Execute command execution
        """
        parser = self._init_parser()
        options = parser.parse_args()
        self._handle(options)


def _get_parser_only() -> argparse.ArgumentParser:
    """
    returns only the initialized parser
    """
    app = TfUtility()
    return app._init_parser()


def main():
    app = TfUtility()
    app.do()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import sys
from datetime import datetime, timedelta

from tfutility.core.base import Command
from tfutility.core.tffile import TfFile
from tfutility.core.tfpaths import TFPaths


class BlockDateHandler(TFPaths, Command):
    block_name = None
    name = None
    help = None  # "Check if a import-blocks has a date comment"

    def add_arguments(self, parser):
        parser = super().add_arguments(parser)

        parser.add_argument(
            "--expire-after",
            type=int,
            help="Returns all blocks that are older than the specified time period. The time span in days",
        )
        parser.add_argument(
            "-s", "--silent", action="store_true", help="Prevent any log output"
        )
        parser.add_argument(
            "--allow-failure",
            action="store_true",
            help="If this flag was set, the module logs only all occurences but does not exit with code 1",
        )
        return parser

    def get_block_name(self):
        if self.block_name is None:
            raise RuntimeError("Please define block_name")
        return self.block_name

    def new_block(self, options, block):
        file_path = block.get_tf_file().path

        dec = block.get_decorator(self.get_name())

        if dec is None:
            self._error = True
            self.get_logger().error(
                f"Missing moveddate Decorator at block '{self.get_block_name()}' in file {file_path} Line {block.start}"
            )
        else:
            now = datetime.now()
            dec_date_create = datetime.strptime(dec.get_parameter("create"), "%d-%m-%Y")

            if not options.expire_after:
                if dec.get_parameter("expire"):
                    dec_date_expire = datetime.strptime(
                        dec.get_parameter("expire"), "%d-%m-%Y"
                    )
                    if now > dec_date_expire:
                        self._error = True
                        self.get_logger().error(
                            f"Moved Block expired in file: {file_path} Line {block.start}"
                        )
            else:
                if now > dec_date_create + timedelta(days=options.expire_after):
                    self._error = True
                    self.get_logger().error(
                        f"Moved Block expired in file: {file_path} Line {block.start}"
                    )

    def handle(self, options):
        self._error = False
        results = super(BlockDateHandler, self).handle(options)

        if not options.allow_failure and self._error:
            sys.exit(1)

        tf_files = self.get_file_list(options.paths)

        for file in tf_files:
            file = TfFile(file)

            for block in file.blocks:
                if block.id.startswith(self.get_block_name()):
                    self.new_block(options, block)

        return results


class ImportDateHandler(BlockDateHandler):
    block_name = "import"
    name = "importdate"
    help = "Check if a import-blocks has a date comment"


class MovedDateHandler(BlockDateHandler):
    block_name = "moved"
    name = "moveddate"
    help = "Check if a moved-blocks has a date comment"



# -*- coding: utf-8 -*-
import argparse

from tfutility.core.abstract import AbstractCommand


class Command(AbstractCommand):
    name = None
    help = None

    def _init(
        self,
        parser: None | argparse.ArgumentParser = None,
        subparser: None | argparse.ArgumentParser = None,
    ) -> argparse.ArgumentParser:
        if parser is None:
            parser = self.parser
        return subparser.add_parser(self.get_name(), help=self.get_help())

    # Overwrite this method to attach Arguments to the existing parser
    def add_arguments(
        self, parser: argparse.ArgumentParser | None
    ) -> argparse.ArgumentParser:
        if parser is None:
            parser = self.parser
        return parser

    def handle(self, options: argparse.Namespace) -> dict:
        return {}



# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

f = open("VERSION", "r")
VERSION = f.read()
f.close()


f = open("README.md", "r")
LONG_DESCRIPTION = f.read()
f.close()

setup(
    name="tfutility",
    version=VERSION,
    description="TfUtility provides different tools for Terraform Developers",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Stefan Eiermann",
    author_email="foss@ultraapp.de",
    url="https://github.com/eieste/tfutility.git",
    license="AGPLv3",
    packages=find_packages(),
    package_dir={"": "src"},
    package_data={"tfutility": ["templates/*"]},
    include_package_data=True,
    entry_points={"console_scripts": ["tfutility = tfutility.main:main"]},
)



# -*- coding: utf-8 -*-
import enum
import re
import sys

from tfutility.core.base import Command
from tfutility.core.tffile import TfFile
from tfutility.core.tfpaths import TFPaths


class SWITCH_DIRECTION(enum.Enum):
    TO_LOCAL = 1
    TO_REMOTE = 2


class SourceSwapHandler(TFPaths, Command):
    name = "sourceswap"
    help = "Allows to switch module Sources. Between a Local and Remote Path"

    def add_arguments(self, parser):
        parser = super().add_arguments(parser)
        parser.add_argument(
            "-s",
            "--switch-to",
            type=str,
            choices=["l", "r", "local", "remote"],
            help="Define to which reference all decorated modules should be swapped",
        )
        return parser

    def block_switch_to(self, options, block, dec, switch_to):
        file_path = block.get_tf_file().path
        if not block.id.startswith("module"):
            self.get_logger().error(
                f"The decorator @sourceswap applied to wrong blocktype in {file_path}:{block.start}"
            )
            sys.exit(1)

        lines = block.get_tf_file().lines
        source_line = -1
        version_line = -1
        source_indent = ""
        for li in range(block.start, block.end):
            cline = lines[li]

            tf_key = re.match(r"(\s*)(\S*)\s*\=\s*.*", cline)
            if tf_key is None:
                continue

            if tf_key.group(2) == "source":
                source_line = li
                source_indent = tf_key.group(1)
            elif tf_key.group(2) == "version":
                version_line = li

        if switch_to is SWITCH_DIRECTION.TO_REMOTE:
            remote_source = dec.get_parameter("remote_source")
            remote_version = dec.get_parameter("remote_version")

            block.get_tf_file().lines[source_line] = re.sub(
                r"source\s*\=\s*\"(.*)\"",
                f'source = "{remote_source}"',
                block.get_tf_file().lines[source_line],
            )

            if version_line == -1:
                block.get_tf_file().lines.insert(
                    source_line, f'{source_indent}version = "{remote_version}"'
                )
            else:
                block.get_tf_file().lines[source_line] = re.sub(
                    r"version\s*\=\s*\"(.*)\"",
                    f'version = "{remote_version}"',
                    block.get_tf_file().lines[source_line],
                )

        else:
            local_source = dec.get_parameter("local_source")
            block.get_tf_file().lines[source_line] = re.sub(
                r"source\s*\=\s*\"(.*)\"",
                f'source = "{local_source}"',
                block.get_tf_file().lines[source_line],
            )

            if version_line > 0:
                del block.get_tf_file().lines[version_line]

    def get_decorator(self, block):
        file_path = block.get_tf_file().path
        dec = block.get_decorator(self.get_name())
        general_error = False
        for param_key in ["remote_source", "remote_version", "local_source"]:
            if not dec.get_parameter(param_key):
                self.get_logger().error(
                    f"Decorator {self.get_name()} {file_path}:{block.start} requires the parameters remote_source, remote_version, local_source"
                )
                general_error = True

        if general_error:
            sys.exit(1)
        return dec

    def handle(self, options):
        if not options.switch_to:
            self.get_logger().error(
                "Please use --switch-to argument with the keywords local or remote"
            )
            sys.exit(1)

        switch_to = SWITCH_DIRECTION.TO_REMOTE
        if options.switch_to in ["local", "l"]:
            switch_to = SWITCH_DIRECTION.TO_LOCAL

        tf_files = self.get_file_list(options.paths)

        for file in tf_files:
            file = TfFile(file)

            for block in file.get_blocks_with_decorator(self.get_name()):
                dec = self.get_decorator(block)
                self.block_switch_to(options, block, dec, switch_to)
            file.write_back()



# -*- coding: utf-8 -*-
import argparse
import logging
import pathlib
import tempfile

import pytest

from tfutility.core.exp import PathIsNotValid
from tfutility.core.tfpaths import TFPaths


def test_get_file_list_invalid_path(caplog):
    parser = argparse.ArgumentParser()

    class Foo:
        def get_logger(self):
            log = logging.getLogger()
            return log

    test = TFPaths(Foo(), parser)
    with pytest.raises(PathIsNotValid, match="Path does not exist"):
        test.get_file_list(
            [
                pathlib.Path("hjklvskjhsdfhkjgsd/"),
            ]
        )

    with tempfile.NamedTemporaryFile(suffix=".tf") as tfile:
        test.get_file_list(
            [
                pathlib.Path(tfile.name),
            ]
        )

    with tempfile.TemporaryDirectory() as tdir:
        test.get_file_list(
            [
                pathlib.Path(tdir),
            ]
        )
        assert "found Zero files" in caplog.text


def test_add_arguments():
    parser = argparse.ArgumentParser()

    class Foo:
        def get_logger(self):
            log = logging.getLogger()
            return log

    TFPaths(Foo(), parser)



# -*- coding: utf-8 -*-
import pathlib
import re
from collections import namedtuple

import hcl2

from tfutility.contrib.deprecation import deprecated

# \#\s?\@([a-z]+)(\((.*)\))?

OpendTfFile = namedtuple("OpendTfFile", ("path", "lines", "parsed"))


class TfUtilityDecorator:
    """
    Represents a tfutility decorator used in tf files above blocks
    """

    # Parses key="value", parameters ( used inside the decorator braces )
    PARAM_REGEX = re.compile(r"(\b\w+)(?:=((?:\"[^\"\\]*(?:\\.[^\"\\]*)*\"|\w+)))?")

    def __init__(self, blockref, name: str, data):
        self._blockref = blockref
        self._name = name
        self._data = data
        self._parameter = None

    def parameter(self, key: str, default=None):
        """
        Returns the value of a given parameter or the default value if the parameter is not set

        :param key: The name of the parameter
        :param default: The default value to return if the parameter is not set
        :return: The value of the parameter or the default value if the parameter is not set
        """
        if self._parameter is None:
            self._parameter = self._parse(self._data)
        return self._parameter.get(key, default)

    @deprecated
    def get_name(self):
        return self._name

    @property
    def name(self):
        return self._name

    def _parse(self, data: str):
        """
        Parses the given data string and returns a dictionary of key-value pairs

        :param data: The data string to parse
        :return: A dictionary of key-value pairs
        """
        result = {}
        for regfind in TfUtilityDecorator.PARAM_REGEX.findall(data):
            result[regfind[0]] = regfind[1].strip('"')
        return result


class TfBlock:
    """
    Represents a block in a terraform file
    """

    # Parses an tfutility decorator inside terraform files
    DECORATOR_REGEX = re.compile(
        r"#\s?\@([a-z]+)(\((.*)\))?",
    )

    def __init__(self, fileref, id: str, block_data: dict):
        """
        Initializes a new instance of the TfBlock class

        :param fileref: The reference to TfFile which contains this block
        :param id: The id of the block. usaly tfblocktype.blocktype.blockname
        :param block_data: The data of the block as parsed by hcl2
        """
        self._fileref = fileref
        self._id = id
        self._decorators = None
        self._content = block_data
        self._start_line = block_data["__start_line__"]
        self._end_line = block_data["__end_line__"]

    @property
    def start(self):
        """
        Start linenumber of the block

        :return: The start linenumber of the block
        :rtype: int
        """
        return self._start_line

    @property
    def end(self):
        """
        End linenumber of the block

        :return: The end linenumber of the block
        :rtype: int
        """
        return self._end_line

    def __str__(self):
        return self._id

    def __repr__(self):
        return f"<BlockWrapper id={self._id}>"

    def __eq__(self, value: object) -> bool:
        self._start_line == value._start_line

    def __lt__(self, other):
        return self.start < other.start

    def __le__(self, other):
        return self.start >= other.start

    def __ne__(self, other):
        return self.start != other.start

    def __gt__(self, other):
        return self.start > other.start

    def __ge__(self, other):
        return self.start >= other.start

    @property
    def id(self):
        """
        id of the block. usually tfblocktype.blocktype.blockname

        :return: The id of the block
        :rtype: str
        """
        return self._id

    @deprecated
    def get_tfile(self):
        return self._fileref.get_tf_file()

    @property
    def tffile(self):
        return self._fileref.get_tf_file()

    @property
    def content(self):
        return self._content

    def get_decorator(self, name: str) -> TfUtilityDecorator | None:
        """
        find a decorator with the given name above the current block

        :param name: The name of the decorator to find
        :type name: str
        :return: The decorator with the given name or None if not found
        :rtype: TfUtilDecorator
        """
        for dec in self.decorators:
            if dec.get_name() == name:
                return dec

    @property
    def decorators(self):
        if self._decorators is None:
            self._decorators = self._find_decorators()
        return self._decorators

    def has_decorator(self, name: str) -> bool:
        """
        Check if this block has a decorator with the given name

        :param name: The name of the decorator to check for
        :type name: str
        :return: True if the block has a decorator with the given name, False otherwise
        :rtype: bool
        """
        if self._decorators is None:
            self._decorators = self._find_decorators()
        for dec in self._decorators:
            if dec.get_name() == name:
                return True
        return False

    def _find_decorators(self):
        """
        search in the file for decorators above this block

        :return: A list of decorators found above this block
        :rtype: list[TfUtilDecorator]
        """
        line_nr = self._start_line - 2
        decorator_list = []
        while self.get_tfile().lines[line_nr].strip().startswith("# @"):
            found_decorator = self.get_tfile().lines[line_nr].strip()
            result = TfBlock.DECORATOR_REGEX.fullmatch(found_decorator)
            decorator_list.append(
                TfUtilityDecorator(self, result.group(1), result.group(2))
            )
            line_nr = -1

        return decorator_list


class TfFile:
    def __init__(self, path: pathlib.Path, autoparse=True):
        self.path = path
        self._blocks = None
        self._tf_file = self.read_tf(path)
        if autoparse:
            self.parse()

    def __repr__(self):
        return f"<TfFile path={self.path.absolute()} >"

    def get_tf_file(self):
        return self._tf_file

    def read_tf(self, path: pathlib.Path):
        with path.open("r") as fobj:
            lines = [line.rstrip() for line in fobj]
            fobj.seek(0)
            return OpendTfFile(path, lines, hcl2.load(fobj, with_meta=True))

    @staticmethod
    def _extend_name(previous_name: str, part: str):
        return ".".join([key for key in previous_name.split(".") + [part] if key])

    @property
    def blocks(self):
        if self._blocks is None:
            self.parse()
        return self._blocks

    def parse(self):
        self._blocks = []
        for block_type, blocks in self._tf_file.parsed.items():
            for block in blocks:
                self._extract_blocks(block, block_type)

    def _extract_blocks(self, blockdata: dict, name: str):
        if isinstance(blockdata, dict) and any(
            [key.startswith("__") for key in blockdata.keys()]
        ):
            elem_names = [key for key in blockdata.keys() if not key.startswith("__")]

            if len(elem_names) > 1:
                new_name = ""
            else:
                new_name = elem_names[0]

            self._blocks.append(
                TfBlock(self, self._extend_name(name, new_name), blockdata)
            )
        else:
            if isinstance(blockdata, list):
                for blockdata in blockdata:
                    self._extract_blocks(blockdata, name)
            else:
                for key, value in blockdata.items():
                    self._extract_blocks(value, self._extend_name(name, key))

    def get_blocks_with_decorator(self, name: str):
        result = []
        for block in self.blocks:
            if block.has_decorator(name):
                result.append(block)
        return result

    def write_back(self):
        with self.path.open("r+") as fobj:
            fobj.truncate()
            for line in self.get_tf_file().lines:
                fobj.write(f"{line}\n")
